
import logging

from gsh.plugin import BaseHostLoader

class MachineLoader(BaseHostLoader):
    opt_short = "-m"
    opt_long = "--machine"
    opt_metavar = "HOST"
    opt_help = "Execute on the specified machine."

    def __call__(self, *args):
        hosts = []
        for machine in args:
            machine = machine.replace(" ", "")
            if "," in machine:
                hosts.extend(machine.split(","))
            else:
                hosts.append(machine)
        return hosts




""" Specify version information about GSH.

This file is meant to be kept minimal. It is loaded by both GSH itself
and setup.py. This is to avoid having to specify the code in multiple
places. Because of this, this file should remain empty other than the
__version__ itself.
"""

__version__ = "0.10.1"




import os
import re

from gsh.plugin import BaseHostLoader
from gsh.exceptions import LoaderError


# These locations will be iterated over using the first
# match to pull hosts from. dsh locations are searched
# to make transition from dsh to gsh more seemless.
GROUP_FILE_LOCATIONS = (
    "~/.gsh/group",
    "~/.dsh/group",
    "/etc/gsh/group",
    "/etc/dsh/group",
)

VALID_GROUPNAME = re.compile(r"^[\w\-=]*$")


class GshLoader(BaseHostLoader):
    opt_short = "-g"
    opt_long = "--group"
    opt_metavar = "GROUP"
    opt_help = "Get a list of machines from a GSH/DSH Group."

    def __call__(self, *args):
        hosts = set()
        groups = set()

        for group in args:
            group = group.replace(" ", "").strip()
            if "," in group:
                groups.update(group.split(","))
            else:
                groups.add(group)

        for group in groups:

            if not VALID_GROUPNAME.match(group):
                raise LoaderError("Invalid group name: %s" % group)

            group_found = False

            for location in GROUP_FILE_LOCATIONS:
                group_filename = os.path.join(os.path.expanduser(location), group)

                if not os.path.isfile(group_filename):
                    continue

                group_found = True
                hosts.update(_read_group_file(group_filename))
                break  # We only want the first match. Stop iteration if we get here.

            if not group_found:
                raise LoaderError("Failed to find group file for: %s" % group)

        return hosts


def _read_group_file(group_filename):
    hosts = set()
    with open(group_filename) as group_file:
        for line in group_file:
            line = line.strip()
            line = line.split("#")[0]
            if not line:
                continue
            hosts.add(line)
    return hosts



import cStringIO
import getpass
import gevent
import select
import socket
import termios
import threading
import tty

from gsh.plugin import BaseExecutor, BaseInnerExecutor


class _ParamikoThread(threading.Thread):
    def __init__(self, executor):
        self.executor = executor
        super(_ParamikoThread, self).__init__()

    def interactive(self, ssh, command):
        stdout = cStringIO.StringIO()
        stderr = cStringIO.StringIO()
        chan = ssh.invoke_shell()
        ssh.get_transport()

        for cmd in command.split("\n"):
            cmd = cmd.strip()
            if not cmd:
                continue
            chan.send(cmd + "\n")

        while True:
            read, write, error = select.select([chan], [], [])
            if chan in read:
                try:
                    out = chan.recv(1024)
                    if not out:
                        break
                    stdout.write(out)
                except socket.timeout():
                    pass

        chan.close()
        stdout.seek(0)
        stderr.seek(0)
        return stdout, stderr


    def run(self):
        # Defer import since most people won't want to use this executor
        # and I don't want it to be required to use gsh.
        import paramiko

        try:
            ssh = paramiko.SSHClient()
            ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
            ssh.connect(self.executor.hostname, password=self.executor.parent.password)

            command = " ".join(self.executor.command)

            if "interactive" in self.executor.parent.args:
                stdout, stderr = self.interactive(ssh, command)
                rv = 0  # Find a way to get rv later.
            else:
                stdin, stdout, stderr = ssh.exec_command(command)
                rv = stdout.channel.recv_exit_status()

            for line in stdout.read().splitlines():
                self.executor.update(self.executor.hostname, "stdout", line)

            for line in stderr.read().splitlines():
                self.executor.update(self.executor.hostname, "stderr", line)

            ssh.close()
            self.executor.rv = rv
        except paramiko.BadAuthenticationType, err:
            self.executor.update(self.executor.hostname, "stderr", "GSH: Failed to login.")
            self.executor.rv = 1


class ParamikoExecutor(BaseExecutor):
    def __init__(self, args, kwargs):

        # This import is unused but I want it to blow up
        # at the job level if it's not found
        import paramiko

        super(ParamikoExecutor, self).__init__(args, kwargs)
        self.password = getpass.getpass("Password: ")
        self.username = getpass.getuser()

    class Executor(BaseInnerExecutor):
        def __init__(self, *args, **kwargs):
            super(ParamikoExecutor.Executor, self).__init__(*args, **kwargs)
            self.rv = 0

        def run(self):

            thread = _ParamikoThread(self)
            thread.daemon = True
            thread.start()

            sleep_duration = 0.01
            while True:
                alive = thread.is_alive()
                if not alive:
                    return self.rv
                gevent.sleep(sleep_duration)
                if sleep_duration < 0.5:
                    sleep_duration *= 2

            return self.rv



""" Configuation related objects and helpers."""

import os
import yaml

from .exceptions import ConfigError


class Config(object):
    """ Configuration object for GSH.

    This object is used as the canonical source for options in GSH. This
    allows for easily setting defaults and having them overridden from
    multiple sources, such as configuration files, arguments on the command
    line and potentially environment variables. This is mostly used for the
    the command line utility.

    Attributes:
        forklimit: The number of concurrent processes to fork at a time.
        print_machines: Whether to prefix output with machine names.
        show_percent: Whether to prefix output with percentage of completion.
        concurrent: Whether to perform operation sequentially vs concurrently.
        timeout: How long to wait for a command to finish on a host.
        plugin_dirs: Where to look for addition plugins.
        hooks: Which hooks to pass through to Gsh.
        executor: Which executor to run the commands with.

    """

    def __init__(self):
        self.forklimit = 64
        self.print_machines = True
        self.show_percent = False
        self.concurrent = True
        self.timeout = 0
        self.plugin_dirs = set()
        self.hooks = set()
        self.executor = "ssh"
        self.executor_args = []
        self.executor_kwargs = {}

    def __repr__(self):
        return (
            "Config(forklimit=%r, print_machines=%r, show_percent=%r, "
            "concurrent=%r, timeout=%r, plugin_dirs=%r, hooks=%r, executor=%r)"
        ) % (
            self.forklimit, self.print_machines, self.show_percent,
            self.concurrent, self.timeout, self.plugin_dirs, self.hooks, self.executor,
        )

    def update_from_file(self, config):
        """ Updates the configuration attributes from a file."""
        try:
            with open(config) as config_file:
                data = yaml.safe_load(config_file)
                if not isinstance(data, dict):
                    data = {}

            self.forklimit = data.get("forklimit", self.forklimit)
            self.print_machines = data.get("print_machines",
                                           self.print_machines)
            self.show_percent = data.get("show_percent",
                                           self.show_percent)
            self.concurrent = data.get("concurrent", self.concurrent)
            self.timeout = data.get("timeout", self.timeout)

            (
                self.executor, self.executor_args, self.executor_kwargs
            ) = self._parse_executor(data.get("executor", self.executor))

            plugin_dirs = data.get("plugin_dirs", [])
            if isinstance(plugin_dirs, basestring):
                plugin_dirs = [plugin_dirs]
            self.plugin_dirs.update(plugin_dirs)

            hooks = data.get("hooks", [])
            if isinstance(hooks, basestring):
                hooks = [hooks]
            self.hooks.update(hooks)

        # It's okay to ignore files that don't exist.
        except IOError:
            pass
        except (yaml.parser.ParserError, yaml.scanner.ScannerError) as err:
            raise ConfigError(
                "Invalid Configuration (%s): %s" % (config, err.problem))

    def _parse_executor(self, executor):
        """ Parses an executor string of the form ssh:arg1,arg2,kwarg1=foo. """

        arguments = ""
        args = []
        kwargs = {}

        if ":" in executor:
            executor, arguments = executor.split(":", 1)

        for argument in arguments.split(","):
            argument = argument.strip()

            if "=" in argument:
                key, value = argument.split("=", 1)
                key = key.strip()
                value = value.strip()
                kwargs[key] = value
            else:
                args.append(argument)

        return executor, args, kwargs


    def update_from_args(self, args):
        """ Update config object from an argparse args object."""

        self.plugin_dirs.update(args.plugin_dirs)
        self.hooks.update(getattr(args, "hooks", []))

        if getattr(args, "forklimit", None) is not None:
            self.forklimit = args.forklimit
        if getattr(args, "print_machines", None) is not None:
            self.print_machines = args.print_machines
        if getattr(args, "show_percent", None) is not None:
            self.show_percent = args.show_percent
        if getattr(args, "concurrent", None) is not None:
            self.concurrent = args.concurrent
        if getattr(args, "timeout", None) is not None:
            self.timeout = args.timeout
        if getattr(args, "executor", None) is not None:
            (
                self.executor, self.executor_args, self.executor_kwargs
            ) = self._parse_executor(args.executor)

    def load_default_files(self):
        """ Update config object from standard config file locations."""
        self.update_from_file("/etc/gsh/gsh.yaml")
        self.update_from_file(os.path.expanduser("~/.gsh/gsh.yaml"))



from core import Gsh
from plugin import get_hooks

def run_remote_command(hosts, command, fork_limit=64, timeout=None):
    hooks = get_hooks()
    buffered_output = hooks.MultiBufferedOutputHook()

    procs = Gsh(
        hosts=hosts,
        command=command,
        fork_limit=fork_limit,
        timeout=timeout,
        hooks=[buffered_output])

    procs.run_async()
    procs.wait()

    return buffered_output.hosts



"""
Module for declaring plugin base classes and helpers.
"""

#pylint: disable=R0921,C0301

import annex
import os

BUILTIN_PLUGIN_DIR = os.path.join(os.path.dirname(os.path.realpath(__file__)), "plugins")


class BaseHostLoader(object):
    """ Abstract Base Class for Writing your own host loaders.
    """

    # If you want this loaded to be available to the gsh
    # command line you can override the following attributes
    # in your plugins.
    opt_short = None
    opt_long = None
    opt_metavar = "FOO"
    opt_help = None

    def __call__(self, *args):
        """ Method called to retreive a list of hosts.

            Args:
                args: A list of keys used to generate a list of hosts.

            Returns:
                A list of hosts.
        """
        raise NotImplementedError()


class BaseExecutionHook(object):
    """ Base class for GSH Hooks.

        This class is meant to be overridden for extending GSH with your
        own hooks. The methods listed below are specified in the order in
        which they will be run.
    """

    def pre_job(self, command, hosts, timestamp):
        """ Called first before any commands are executed.

            Args:
                command: The command line that will be passed to each host.
                hosts: A list of hosts that the command will be executed on.
                timestamp: A Unix timestamp when this hook was run.
        """

    def pre_host(self, hostname, timestamp):
        """ Called for each host, before a host has started executing a command.

            Args:
                hostname: The host we're about to run a command on.
                timestamp: A Unix timestamp when this hook was run.
        """

    def update_host(self, hostname, stream, line):
        """ Called each time a new line of text has been streamed from a host.

            Args:
                hostname: The host where the output has come from.
                stream: Which output stream generated the output (stdout/stderr.)
                line: The line of output streamed from the host.
        """

    def post_host(self, hostname, return_code, timestamp):
        """ Called for each host, after a host has finished executing a command.

            Args:
                hostname: The host we're about to run a command on.
                return_code: The exitcode received from the remote command.
                timestamp: A Unix timestamp when this hook was run.
        """

    def post_job(self, timestamp):
        """ Called last after all commands have been executed.

            Args:
                timestamp: A Unix timestamp when this hook was run.
        """


class BaseExecutor(object):
    """ Base Executor to be instantiated at the job level.

        Your TopLevel Base Executor will define an inner class
        called Executor, which inherits from BaseInnerExecutor,
        which will be instantiated per host. That is where most of
        your logic will go. This outer class is for setting up
        attributes at the job level.

    Attributes:
        args: list of arguments passed through.
        kwargs: dict of keyword arguments pass through.
    """

    def __init__(self, args=None, kwargs=None):
        self.args = [] if args is None else args
        self.kwargs = {} if kwargs is None else kwargs


class BaseInnerExecutor(object):
    """ Executor to be instantiated per host.

    Attributes:
        parent: The outer, job-level, executor object.
        hostname: The hostname you'll be executing on.
        command: The command to run on the host.
        timeout: Max time before failure.
        update: A callback to write the output from the command.

    """

    def __init__(self, parent, hostname, command, timeout, update):
        self.parent = parent
        self.hostname = hostname
        self.command = command
        self.timeout = timeout
        self.update = update

    def run(self):
        """ The actual execution logic goes here.

            Returns:
                An integer, the return code of the command.
        """
        return 0


def get_loaders(additional_dirs=None):
    """ Helper function to find and load all loaders. """
    if additional_dirs is None:
        additional_dirs = []
    return annex.Annex(BaseHostLoader, [
        os.path.join(BUILTIN_PLUGIN_DIR, "loaders"),
        "/etc/gsh/plugins/loaders",
        [os.path.join(plugin_dir, "loaders") for plugin_dir in additional_dirs]
    ])


def get_hooks(additional_dirs=None):
    """ Helper function to find and load all hooks. """
    if additional_dirs is None:
        additional_dirs = []
    return annex.Annex(BaseExecutionHook, [
        os.path.join(BUILTIN_PLUGIN_DIR, "hooks"),
        "/etc/gsh/plugins/hooks",
        [os.path.join(plugin_dir, "hooks") for plugin_dir in additional_dirs]
    ], instantiate=False)


def get_executors(additional_dirs=None):
    """ Helper function to find and load all executors. """
    if additional_dirs is None:
        additional_dirs = []
    return annex.Annex(BaseExecutor, [
        os.path.join(BUILTIN_PLUGIN_DIR, "executors"),
        "/etc/gsh/plugins/executors",
        [os.path.join(plugin_dir, "executors") for plugin_dir in additional_dirs]
    ], instantiate=False)



#!/usr/bin/env python

import setuptools
from distutils.core import setup

execfile('gsh/version.py')

with open('requirements.txt') as requirements:
    required = requirements.read().splitlines()

kwargs = {
    "name": "gsh",
    "version": str(__version__),
    "packages": ["gsh"],
    "package_data": {"gsh": [
        "plugins/loaders/*.py",
        "plugins/hooks/*.py",
        "plugins/executors/*.py",
    ]},
    "scripts": ["bin/gsh"],
    "description": "Pluggable Distributed SSH Command Executer.",
    # PyPi, despite not parsing markdown, will prefer the README.md to the
    # standard README. Explicitly read it here.
    "long_description": open("README").read(),
    "author": "Gary M. Josack",
    "maintainer": "Gary M. Josack",
    "author_email": "gary@byoteki.com",
    "maintainer_email": "gary@byoteki.com",
    "license": "MIT",
    "install_requires": required,
    "url": "https://github.com/gmjosack/gsh",
    "download_url": "https://github.com/gmjosack/gsh/archive/master.tar.gz",
    "classifiers": [
        "Programming Language :: Python",
        "Topic :: Software Development",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ]
}

setup(**kwargs)



from gsh.plugin import BaseExecutionHook

class BufferedOutputHook(BaseExecutionHook):
    """ Hook to capture output from a RemotePopen call."""
    def __init__(self):
        self.stdout = []
        self.stderr = []
        self.rc = None

    def update_host(self, hostname, stream, line):
        if stream not in ("stdout", "stderr"):
            return
        getattr(self, stream).append(line)

    def post_host(self, hostname, rc, timestamp):
        self.rc = rc


class MultiBufferedOutputHook(BaseExecutionHook):
    """ Hook to gather output from all hosts on a Gsh call."""
    def __init__(self):
        self.hosts = {}

    def pre_host(self, hostname, timestamp):
        self.hosts[hostname] = BufferedOutputHook()

    def update_host(self, hostname, *args, **kwargs):
        self.hosts[hostname].update_host(hostname, *args, **kwargs)

    def post_host(self, hostname, *args, **kwargs):
        self.hosts[hostname].post_host(hostname, *args, **kwargs)


